import casadi as cas
from typing import Union, List
import numpy as np
import pytest


class Opti(cas.Opti):
    def __init__(self,
                 variable_categories_to_freeze=[],
                 ):
        super().__init__()
        p_opts = {}
        s_opts = {}
        s_opts["max_iter"] = 3000
        s_opts["mu_strategy"] = "adaptive"
        self.solver('ipopt', p_opts, s_opts)  # Default to IPOPT solver
        self.categories_to_freeze = variable_categories_to_freeze

    def variable(self,
                 n_vars: int = 1,
                 init_guess: Union[float, np.ndarray] = None,
                 scale: float = None,
                 log_transform: bool = False,
                 category: str = "Default",
                 freeze: bool = False,
                 ) -> cas.MX:
        """
        Initializes a new decision variable.
        
        It is recommended that you provide an initial guess (`init_guess`) and scale (`scale`) for each variable,
        although these are not strictly required.

        Args:
            n_vars: Number of variables to initialize (used to initialize a vector of variables). If you are
                initializing a scalar variable (the most typical case), leave this equal to 1. When using vector variables,
                individual components of this vector of variables can be accessed via indexing.

                Example:
                    >>> opti = asb.Opti()
                    >>> my_var = opti.variable(n_vars = 5)
                    >>> opti.subject_to(my_var[3] >= my_var[2])  # This is a valid way of indexing
                    >>> my_sum = asb.cas.sum1(my_var)  # This will sum up all elements of `my_var`

            init_guess: Initial guess for the variable being initialized. For scalar variables, this should be a
                float. For vector variables (see `n_vars`), you can provide either a float (in which case all elements
                of the vector will be initialized to the given value) or an iterable of equal length (in which case
                each element will be initialized to the corresponding value in the given iterable).

                In the case where the variable is to be log-transformed (see `log_transform`), the initial guess should
                not be log-transformed as well; this happens under the hood. The initial guess must, of course, be a
                positive number in this case.

                If not specified, initial guess defaults to 0 for non-log-transformed variables and 1 for
                log-transformed variables.

            scale: Approximate scale of the variable 

            log_transform:
            category:

        Returns:
            The variable itself as a symbolic CasADi variable (MX type).

        """
        # Validate the inputs
        if log_transform and init_guess is not None:
            if np.any(init_guess <= 0):
                raise ValueError(
                    "If you are initializing a log-transformed variable, the initial guess(es) must be positive.")

        # Set defaults
        if init_guess is None:
            init_guess = 1 if log_transform else 0
        if scale is None:
            scale = init_guess if log_transform else 1

        # Validate the inputs
        if np.any(scale <= 0):
            raise ValueError("The 'scale' argument must be a positive number.")

        # If the variable is in a category to be frozen, fix the variable at the initial guess.
        if category in self.categories_to_freeze:
            freeze = True

        # If the variable is to be frozen, return the initial guess. Otherwise, define the variable using CasADi symbolics.
        if freeze:
            var = init_guess * np.ones(n_vars)
        else:
            if not log_transform:
                var = scale * super().variable(n_vars)
                self.set_initial(var, init_guess)
            else:
                log_scale = scale / init_guess
                log_var = log_scale * super().variable(n_vars)
                var = cas.exp(log_var)
                self.set_initial(log_var, cas.log(init_guess))

        return var

    def subject_to(self,
                   constraint: Union[cas.MX, bool, List],
                   ) -> None:
        # Determine whether you're dealing with a single (possibly vectorized) constraint or a list of constraints.
        # If the latter, recursively apply them.
        if isinstance(constraint, List):
            for each_constraint in constraint:
                self.subject_to(each_constraint)
            return

        # If it's a proper constraint (MX type), pass it into the problem formulation and be done with it.
        if isinstance(constraint, cas.MX):
            super().subject_to(constraint)
            return

        # If the constraint(s) always evaluates True (e.g. if you enter "5 > 3"), skip it.
        # This allows you to toggle frozen variables without causing problems with setting up constraints.
        elif np.all(constraint):
            pass

        # If any of the constraint(s) are always False (e.g. if you enter "5 < 3"), raise an error.
        # This indicates that the problem is infeasible as-written, likely because the user has frozen too
        # many decision variables using the Opti.variable(freeze=True) syntax.
        elif np.any(np.logical_not(constraint)):
            raise RuntimeError(f"""The problem is infeasible due to a constraint that always evaluates False. You 
            supplied the following constraint: {constraint}. This can happen if you've frozen too 
            many decision variables, leading to an overconstrained problem.""")

        else:  # In theory, this should never be called, so long as the constraints can be boolean-evaluated.
            raise TypeError(f"""Opti.subject_to could not determine the truthiness of your constraint, and it
            doesn't appear to be a symbolic type or a boolean type. You supplied the following constraint:
            {constraint}""")


if __name__ == '__main__':
    pytest.main()



from aerosandbox.geometry import *
from aerosandbox.performance import *


class AeroProblem(AeroSandboxObject):
    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 ):
        self.airplane = airplane
        self.op_point = op_point



import aerosandbox as asb
from aerosandbox import cas
import numpy as np
import pytest

"""
These tests all implement a simple wing aerostructural optimization problem.

Problem is taken from Section 3 of "Geometric Programming for Aircraft Design Optimization" by W. Hoburg and P. Abbeel.
http://web.mit.edu/~whoburg/www/papers/hoburgabbeel2014.pdf

GPKit implementation available at: https://gpkit.readthedocs.io/en/latest/examples.html#simple-wing

Each test solves the same problem in a slightly different way, verifying the correct solution is obtained for all flags.

"""
from numpy import pi

### Constants
k = 1.2  # form factor [-]
e = 0.95  # Oswald efficiency factor [-]
mu = 1.78e-5  # viscosity of air [kg/m/s]
rho = 1.23  # density of air [kg/m^3]
tau = 0.12  # airfoil thickness to chord ratio [-]
N_ult = 3.8  # ultimate load factor [-]
V_min = 22  # takeoff speed [m/s]
C_Lmax = 1.5  # max CL with flaps down [-]
S_wetratio = 2.05  # wetted area ratio [-]
W_W_coeff1 = 8.71e-5  # Wing Weight Coefficient 1 [1/m]
W_W_coeff2 = 45.24  # Wing Weight Coefficient 2 [Pa]
CDA0 = 0.031  # fuselage drag area [m^2]
W_0 = 4940.0  # aircraft weight excluding wing [N]


def test_original_gpkit_like_solve():
    opti = asb.Opti()  # initialize an optimization environment

    ### Variables
    D = opti.variable(log_transform=True, init_guess=1000, scale=1e3)  # total drag force [N]
    A = opti.variable(log_transform=True, init_guess=10, scale=1e1)  # aspect ratio
    S = opti.variable(log_transform=True, init_guess=100, scale=1e2)  # total wing area [m^2]
    V = opti.variable(log_transform=True, init_guess=100, scale=1e2)  # cruising speed [m/s]
    W = opti.variable(log_transform=True, init_guess=8e3, scale=1e4)  # total aircraft weight [N]
    Re = opti.variable(log_transform=True, init_guess=5e6, scale=1e6)  # Reynolds number [-]
    C_D = opti.variable(log_transform=True, init_guess=0.03, scale=1e-2)  # Drag coefficient of wing [-]
    C_L = opti.variable(log_transform=True, init_guess=1, scale=1e-1)  # Lift coefficient of wing [-]
    C_f = opti.variable(log_transform=True, init_guess=0.01, scale=1e-2)  # Skin friction coefficient [-]
    W_w = opti.variable(log_transform=True, init_guess=3e3, scale=1e3)  # Wing weight [N]

    ### Constraints
    # Drag model
    C_D_fuse = CDA0 / S
    C_D_wpar = k * C_f * S_wetratio
    C_D_ind = C_L ** 2 / (pi * A * e)
    opti.subject_to(cas.log(C_D) >= cas.log(C_D_fuse + C_D_wpar + C_D_ind))

    # Wing weight model
    W_w_strc = W_W_coeff1 * (N_ult * A ** 1.5 * (W_0 * W * S) ** 0.5) / tau
    W_w_surf = W_W_coeff2 * S
    opti.subject_to(cas.log(W_w) >= cas.log(W_w_surf + W_w_strc))

    # Other models
    opti.subject_to([
        cas.log(D) >= cas.log(0.5 * rho * S * C_D * V ** 2),
        cas.log(Re) <= cas.log((rho / mu) * V * (S / A) ** 0.5),
        cas.log(C_f) >= cas.log(0.074 / Re ** 0.2),
        cas.log(W) <= cas.log(0.5 * rho * S * C_L * V ** 2),
        cas.log(W) <= cas.log(0.5 * rho * S * C_Lmax * V_min ** 2),
        cas.log(W) >= cas.log(W_0 + W_w),
    ])

    # Objective
    opti.minimize(cas.log(D))

    sol = opti.solve()

    assert sol.value(D) == pytest.approx(303.1, abs=0.1)


def test_rewritten_gpkit_like_solve():
    opti = asb.Opti()  # initialize an optimization environment

    ### Variables
    A = opti.variable(log_transform=True, init_guess=10, scale=1e1)  # aspect ratio
    S = opti.variable(log_transform=True, init_guess=100, scale=1e2)  # total wing area [m^2]
    V = opti.variable(log_transform=True, init_guess=100, scale=1e2)  # cruising speed [m/s]
    W = opti.variable(log_transform=True, init_guess=8e3, scale=1e4)  # total aircraft weight [N]
    C_L = opti.variable(log_transform=True, init_guess=1, scale=1e-1)  # Lift coefficient of wing [-]

    ### Constraints
    # Aerodynamics model
    C_D_fuse = CDA0 / S
    Re = (rho / mu) * V * (S / A) ** 0.5
    C_f = 0.074 / Re ** 0.2
    C_D_wpar = k * C_f * S_wetratio
    C_D_ind = C_L ** 2 / (pi * A * e)
    C_D = C_D_fuse + C_D_wpar + C_D_ind
    q = 0.5 * rho * V ** 2
    D = q * S * C_D
    L_cruise = q * S * C_L
    L_takeoff = 0.5 * rho * S * C_Lmax * V_min ** 2

    # Wing weight model
    W_w_strc = W_W_coeff1 * (N_ult * A ** 1.5 * (W_0 * W * S) ** 0.5) / tau
    W_w_surf = W_W_coeff2 * S
    W_w = W_w_surf + W_w_strc

    # Other constraints
    opti.subject_to([
        W <= L_cruise,
        W <= L_takeoff,
        W >= W_0 + W_w
    ])

    # Objective
    opti.minimize(D)

    sol = opti.solve()

    assert sol.value(D) == pytest.approx(303.1, abs=0.1)


def test_rewritten_gpkit_like_solve_no_guesses_or_scaling():
    opti = asb.Opti()  # initialize an optimization environment

    ### Variables
    A = opti.variable(log_transform=True)  # aspect ratio
    S = opti.variable(log_transform=True)  # total wing area [m^2]
    V = opti.variable(log_transform=True)  # cruising speed [m/s]
    W = opti.variable(log_transform=True)  # total aircraft weight [N]
    C_L = opti.variable(log_transform=True)  # Lift coefficient of wing [-]

    ### Constraints
    # Aerodynamics model
    C_D_fuse = CDA0 / S
    Re = (rho / mu) * V * (S / A) ** 0.5
    C_f = 0.074 / Re ** 0.2
    C_D_wpar = k * C_f * S_wetratio
    C_D_ind = C_L ** 2 / (pi * A * e)
    C_D = C_D_fuse + C_D_wpar + C_D_ind
    q = 0.5 * rho * V ** 2
    D = q * S * C_D
    L_cruise = q * S * C_L
    L_takeoff = 0.5 * rho * S * C_Lmax * V_min ** 2

    # Wing weight model
    W_w_strc = W_W_coeff1 * (N_ult * A ** 1.5 * (W_0 * W * S) ** 0.5) / tau
    W_w_surf = W_W_coeff2 * S
    W_w = W_w_surf + W_w_strc

    # Other constraints
    opti.subject_to([
        W <= L_cruise,
        W <= L_takeoff,
        W == W_0 + W_w
    ])

    # Objective
    opti.minimize(D)

    sol = opti.solve()

    assert sol.value(D) == pytest.approx(303.1, abs=0.1)


def test_non_log_transformed_solve():
    opti = asb.Opti()  # initialize an optimization environment

    ### Variables
    A = opti.variable(init_guess=10, scale=1e1)  # aspect ratio
    S = opti.variable(init_guess=100, scale=1e2)  # total wing area [m^2]
    V = opti.variable(init_guess=100, scale=1e2)  # cruising speed [m/s]
    W = opti.variable(init_guess=8e3, scale=1e4)  # total aircraft weight [N]
    C_L = opti.variable(init_guess=1, scale=1e-1)  # Lift coefficient of wing [-]

    ### Constraints
    # Aerodynamics model
    C_D_fuse = CDA0 / S
    Re = (rho / mu) * V * (S / A) ** 0.5
    C_f = 0.074 / Re ** 0.2
    C_D_wpar = k * C_f * S_wetratio
    C_D_ind = C_L ** 2 / (pi * A * e)
    C_D = C_D_fuse + C_D_wpar + C_D_ind
    q = 0.5 * rho * V ** 2
    D = q * S * C_D
    L_cruise = q * S * C_L
    L_takeoff = 0.5 * rho * S * C_Lmax * V_min ** 2

    # Wing weight model
    W_w_strc = W_W_coeff1 * (N_ult * A ** 1.5 * (W_0 * W * S) ** 0.5) / tau
    W_w_surf = W_W_coeff2 * S
    W_w = W_w_surf + W_w_strc

    # Other constraints
    opti.subject_to([
        W <= L_cruise,
        W <= L_takeoff,
        W == W_0 + W_w
    ])

    # Objective
    opti.minimize(D)

    sol = opti.solve()

    assert sol.value(D) == pytest.approx(303.1, abs=0.1)


if __name__ == '__main__':
    pytest.main()



from aerosandbox_legacy_v0 import *

a = Airplane(
        name="Conventional",
        xyz_ref=[0.05, 0, 0],
        wings=[
            Wing(
                name="Main Wing",
                xyz_le=[0, 0, 0],
                symmetric=True,
                xsecs=[
                    WingXSec(  # Root
                        xyz_le=[0, 0, 0],
                        chord=0.18,
                        twist=2,
                        airfoil=Airfoil(name="naca4412")
                    ),
                    WingXSec(  # Mid
                        xyz_le=[0.01, 0.5, 0],
                        chord=0.16,
                        twist=0,
                        airfoil=Airfoil(name="naca4412")
                    ),
                    WingXSec(  # Tip
                        xyz_le=[0.08, 1, 0.1],
                        chord=0.08,
                        twist=-2,
                        airfoil=Airfoil(name="naca4412")
                    )
                ]
            ),
            Wing(
                name="Horizontal Stabilizer",
                xyz_le=[0.6, 0, 0.1],
                symmetric=True,
                xsecs=[
                    WingXSec(  # root
                        xyz_le=[0, 0, 0],
                        chord=0.1,
                        twist=-10,
                        airfoil=Airfoil(name="naca0012")
                    ),
                    WingXSec(  # tip
                        xyz_le=[0.02, 0.17, 0],
                        chord=0.08,
                        twist=-10,
                        airfoil=Airfoil(name="naca0012")
                    )
                ]
            ),
            Wing(
                name="Vertical Stabilizer",
                xyz_le=[0.6, 0, 0.15],
                symmetric=False,
                xsecs=[
                    WingXSec(
                        xyz_le=[0, 0, 0],
                        chord=0.1,
                        twist=0,
                        airfoil=Airfoil(name="naca0012")
                    ),
                    WingXSec(
                        xyz_le=[0.04, 0, 0.15],
                        chord=0.06,
                        twist=0,
                        airfoil=Airfoil(name="naca0012")
                    )
                ]
            )
        ]
    )
a.set_ref_dims_from_wing()

a.draw()


from aerosandbox.geometry.common import *
from typing import List


class Wing(AeroSandboxObject):
    """
    Definition for a wing.
    If the wing is symmetric across the XZ plane, just define the right half and supply "symmetric = True" in the constructor.
    If the wing is not symmetric across the XZ plane, just define the wing.
    """

    def __init__(self,
                 name: str = "Untitled Wing",  # It can help when debugging to give each wing a sensible name.
                 x_le: float = 0,  # Will translate all of the xsecs of the wing. Useful for moving the wing around.
                 y_le: float = 0,  # Will translate all of the xsecs of the wing. Useful for moving the wing around.
                 z_le: float = 0,  # Will translate all of the xsecs of the wing. Useful for moving the wing around.
                 xsecs: List['WingXSec'] = [],  # This should be a list of WingXSec objects.
                 symmetric: bool = False,  # Is the wing symmetric across the XZ plane?
                 chordwise_panels: int = 8,  # The number of chordwise panels to be used in VLM and panel analyses.
                 chordwise_spacing: str = "cosine",  # Can be 'cosine' or 'uniform'. Highly recommended to be cosine.
                 ):
        self.name = name
        self.xyz_le = cas.vertcat(x_le, y_le, z_le)
        self.xsecs = xsecs
        self.symmetric = symmetric
        self.chordwise_panels = chordwise_panels
        self.chordwise_spacing = chordwise_spacing

    def __repr__(self) -> str:
        return "Wing %s (%i xsecs, %s)" % (
            self.name,
            len(self.xsecs),
            "symmetric" if self.symmetric else "not symmetric"
        )

    def area(self,
             type: str = "wetted"
             ) -> float:
        """
        Returns the area, with options for various ways of measuring this.
         * wetted: wetted area
         * projected: area projected onto the XY plane (top-down view)
        :param type:
        :return:
        """
        area = 0
        for i in range(len(self.xsecs) - 1):
            chord_eff = (self.xsecs[i].chord
                         + self.xsecs[i + 1].chord) / 2
            this_xyz_te = self.xsecs[i].xyz_te()
            that_xyz_te = self.xsecs[i + 1].xyz_te()
            if type == "wetted":
                span_le_eff = cas.sqrt(
                    (self.xsecs[i].xyz_le[1] - self.xsecs[i + 1].xyz_le[1]) ** 2 +
                    (self.xsecs[i].xyz_le[2] - self.xsecs[i + 1].xyz_le[2]) ** 2
                )
                span_te_eff = cas.sqrt(
                    (this_xyz_te[1] - that_xyz_te[1]) ** 2 +
                    (this_xyz_te[2] - that_xyz_te[2]) ** 2
                )
            elif type == "projected":
                span_le_eff = cas.fabs(
                    self.xsecs[i].xyz_le[1] - self.xsecs[i + 1].xyz_le[1]
                )
                span_te_eff = cas.fabs(
                    this_xyz_te[1] - that_xyz_te[1]
                )
            else:
                raise ValueError("Bad value of 'type'!")

            span_eff = (span_le_eff + span_te_eff) / 2
            area += chord_eff * span_eff
        if self.symmetric:
            area *= 2
        return area

    def span(self,
             type: str = "wetted"
             ) -> float:
        """
        Returns the span, with options for various ways of measuring this.
         * wetted: Adds up YZ-distances of each section piece by piece
         * yz: YZ-distance between the root and tip of the wing
         * y: Y-distance between the root and tip of the wing
         * z: Z-distance between the root and tip of the wing
        If symmetric, this is doubled to obtain the full span.
        :param type: One of the above options, as a string.
        :return: span
        """
        if type == "wetted":
            span = 0
            for i in range(len(self.xsecs) - 1):
                sect1_xyz_le = self.xsecs[i].xyz_le
                sect2_xyz_le = self.xsecs[i + 1].xyz_le
                sect1_xyz_te = self.xsecs[i].xyz_te()
                sect2_xyz_te = self.xsecs[i + 1].xyz_te()

                span_le = cas.sqrt(
                    (sect1_xyz_le[1] - sect2_xyz_le[1]) ** 2 +
                    (sect1_xyz_le[2] - sect2_xyz_le[2]) ** 2
                )
                span_te = cas.sqrt(
                    (sect1_xyz_te[1] - sect2_xyz_te[1]) ** 2 +
                    (sect1_xyz_te[2] - sect2_xyz_te[2]) ** 2
                )
                span_eff = (span_le + span_te) / 2
                span += span_eff

        elif type == "yz":
            root = self.xsecs[0]  # type: WingXSec
            tip = self.xsecs[-1]  # type: WingXSec
            span = cas.sqrt(
                (root.xyz_le[1] - tip.xyz_le[1]) ** 2 +
                (root.xyz_le[2] - tip.xyz_le[2]) ** 2
            )
        elif type == "y":
            root = self.xsecs[0]  # type: WingXSec
            tip = self.xsecs[-1]  # type: WingXSec
            span = cas.fabs(
                tip.xyz_le[1] - root.xyz_le[1]
            )
        elif type == "z":
            root = self.xsecs[0]  # type: WingXSec
            tip = self.xsecs[-1]  # type: WingXSec
            span = cas.fabs(
                tip.xyz_le[2] - root.xyz_le[2]
            )
        else:
            raise ValueError("Bad value of 'type'!")
        if self.symmetric:
            span *= 2
        return span

    def aspect_ratio(self) -> float:
        # Returns the aspect ratio (b^2/S).
        # Uses the full span and the full area if symmetric.
        return self.span() ** 2 / self.area()

    def has_symmetric_control_surfaces(self) -> bool:
        # Returns a boolean of whether the wing is totally symmetric (i.e.), every xsec has control_surface_type = "symmetric".
        for xsec in self.xsecs:
            if not xsec.control_surface_type == "symmetric":
                return False
        return True

    def mean_geometric_chord(self) -> float:
        """
        Returns the mean geometric chord of the wing (S/b).
        :return:
        """
        return self.area() / self.span()

    def mean_aerodynamic_chord(self) -> float:
        """
        Returns the mean aerodynamic chord of the wing.
        """
        area = 0
        sum_dMAC_dA = 0

        for i, xsec_a in enumerate(self.xsecs[:-1]):
            xsec_b = self.xsecs[i + 1]

            section_span = cas.sqrt(
                cas.sumsqr(xsec_a.quarter_chord()[1:, :] - xsec_b.quarter_chord()[1:, :])
            )

            section_taper_ratio = xsec_b.chord / xsec_a.chord
            section_area = (xsec_a.chord + xsec_b.chord) / 2 * section_span
            section_MAC = (2 / 3) * xsec_a.chord * (
                    (1 + section_taper_ratio + section_taper_ratio ** 2) /
                    (1 + section_taper_ratio)
            )
            area = area + section_area
            sum_dMAC_dA = sum_dMAC_dA + section_area * section_MAC

        MAC = sum_dMAC_dA / area

        return MAC

    # def mean_aerodynamic_chord_location(self): # TODO verify and add
    #     """
    #     Returns the x and y LE position of the mean aerodynamic chord.
    #     Based on the same assumptions as wing.mean_aerodynamic_chord.
    #     """
    #     area = 0
    #     sum_dx_dA = 0
    #     sum_dy_dA = 0
    #     sum_dz_dA = 0
    #
    #     for i, xsec_a in enumerate(self.xsecs[:-1]):
    #         xsec_b = self.xsecs[i + 1]
    #
    #         c_r = xsec_a.chord
    #         c_t = xsec_b.chord
    #         x_r = xsec_a.x_le
    #         x_t = xsec_b.x_le
    #         y_r = xsec_a.y_le
    #         y_t = xsec_b.y_le
    #         z_r = xsec_a.z_le
    #         z_t = xsec_b.z_le
    #
    #         taper_ratio = c_t / c_r
    #         d_area = (c_r + c_t) * (y_t - y_r) / 2
    #         area = area + d_area
    #
    #         x_mac = x_r + (x_t - x_r) * (1 + 2 * taper_ratio) / (3 + 3 * taper_ratio)
    #         sum_dx_dA = d_area * x_mac + sum_dx_dA
    #         y_mac = y_r + (y_t - y_r) * (1 + 2 * taper_ratio) / (3 + 3 * taper_ratio)
    #         sum_dy_dA = d_area * y_mac + sum_dy_dA
    #         z_mac = z_r + (z_t - z_r) * (1 + 2 * taper_ratio) / (3 + 3 * taper_ratio)
    #         sum_dz_dA = d_area * z_mac + sum_dz_dA
    #
    #     x_mac = sum_dx_dA / area
    #     y_mac = sum_dy_dA / area
    #     z_mac = sum_dz_dA / area
    #
    #     return cas.vertcat(x_mac, y_mac, z_mac)

    def mean_twist_angle(self) -> float:
        """
        Returns the mean twist angle (in degrees) of the wing, weighted by span.
        You can think of it as \int_{b}(twist)db, where b is span.
        WARNING: This function's output is only exact in the case where all of the cross sections have the same twist axis!
        :return: mean twist angle (in degrees)
        """
        # First, find the spans
        span = []
        for i in range(len(self.xsecs) - 1):
            sect1_xyz_le = self.xsecs[i].xyz_le
            sect2_xyz_le = self.xsecs[i + 1].xyz_le
            sect1_xyz_te = self.xsecs[i].xyz_te()
            sect2_xyz_te = self.xsecs[i + 1].xyz_te()

            span_le = cas.sqrt(
                (sect1_xyz_le[1] - sect2_xyz_le[1]) ** 2 +
                (sect1_xyz_le[2] - sect2_xyz_le[2]) ** 2
            )
            span_te = cas.sqrt(
                (sect1_xyz_te[1] - sect2_xyz_te[1]) ** 2 +
                (sect1_xyz_te[2] - sect2_xyz_te[2]) ** 2
            )
            span_eff = (span_le + span_te) / 2
            span.append(span_eff)

        # Then, find the twist-span product
        twist_span_product = 0
        for i in range(len(self.xsecs)):
            xsec = self.xsecs[i]
            if i > 0:
                twist_span_product += xsec.twist * span[i - 1] / 2
            if i < len(self.xsecs) - 1:
                twist_span_product += xsec.twist * span[i] / 2

        # Then, divide
        mean_twist = twist_span_product / cas.sum1(cas.vertcat(*span))
        return mean_twist

    def mean_sweep_angle(self) -> float:
        """
        Returns the mean quarter-chord sweep angle (in degrees) of the wing, relative to the x-axis.
        Positive sweep is backwards, negative sweep is forward.
        :return:
        """
        root_quarter_chord = 0.75 * self.xsecs[0].xyz_le + 0.25 * self.xsecs[0].xyz_te()
        tip_quarter_chord = 0.75 * self.xsecs[-1].xyz_le + 0.25 * self.xsecs[-1].xyz_te()

        vec = tip_quarter_chord - root_quarter_chord
        vec_norm = vec / cas.norm_2(vec)

        sin_sweep = vec_norm[0]  # from dot product with x_hat

        sweep_deg = cas.asin(sin_sweep) * 180 / cas.pi

        return sweep_deg

    def approximate_center_of_pressure(self) -> cas.DM:
        """
        Returns the approximate location of the center of pressure. Given as the area-weighted quarter chord of the wing.
        :return: [x, y, z] of the approximate center of pressure
        """
        areas = []
        quarter_chord_centroids = []
        for i in range(len(self.xsecs) - 1):
            # Find areas
            chord_eff = (self.xsecs[i].chord
                         + self.xsecs[i + 1].chord) / 2
            this_xyz_te = self.xsecs[i].xyz_te()
            that_xyz_te = self.xsecs[i + 1].xyz_te()
            span_le_eff = cas.sqrt(
                (self.xsecs[i].xyz_le[1] - self.xsecs[i + 1].xyz_le[1]) ** 2 +
                (self.xsecs[i].xyz_le[2] - self.xsecs[i + 1].xyz_le[2]) ** 2
            )
            span_te_eff = cas.sqrt(
                (this_xyz_te[1] - that_xyz_te[1]) ** 2 +
                (this_xyz_te[2] - that_xyz_te[2]) ** 2
            )
            span_eff = (span_le_eff + span_te_eff) / 2

            areas.append(chord_eff * span_eff)

            # Find quarter-chord centroids of each section
            quarter_chord_centroids.append(
                (
                        0.75 * self.xsecs[i].xyz_le + 0.25 * self.xsecs[i].xyz_te() +
                        0.75 * self.xsecs[i + 1].xyz_le + 0.25 * self.xsecs[i + 1].xyz_te()
                ) / 2 + self.xyz_le
            )

        areas = cas.vertcat(*areas)
        quarter_chord_centroids = cas.transpose(cas.horzcat(*quarter_chord_centroids))

        total_area = cas.sum1(areas)
        approximate_cop = cas.sum1(areas / cas.sum1(areas) * quarter_chord_centroids)

        if self.symmetric:
            approximate_cop[:, 1] = 0

        return approximate_cop

    def taper_ratio(self) -> float:
        """
        Gives the taper ratio of the Wing. Strictly speaking, only valid for trapezoidal wings.

        Returns:
            Taper ratio of the Wing.

        """
        return self.xsecs[-1].chord / self.xsecs[0].chord


class WingXSec(AeroSandboxObject):
    """
    Definition for a wing cross section ("X-section").
    """

    def __init__(self,
                 x_le: float = 0,  # Coordinate of the leading edge of the cross section, relative to the wing's datum.
                 y_le: float = 0,  # Coordinate of the leading edge of the cross section, relative to the wing's datum.
                 z_le: float = 0,  # Coordinate of the leading edge of the cross section, relative to the wing's datum.
                 chord: float = 0,  # Chord of the wing at this cross section
                 twist: float = 0,  # Twist always occurs about the leading edge!
                 twist_axis: cas.DM = cas.DM([0, 1, 0]),  # By default, always twists about the Y-axis.
                 airfoil: 'Airfoil' = None,  # The airfoil to be used at this cross section.
                 control_surface_type: str = "symmetric",  # TODO change this to a boolean
                 # Can be "symmetric" or "asymmetric". Symmetric is like flaps, asymmetric is like an aileron.
                 control_surface_hinge_point: float = 0.75,
                 # The location of the control surface hinge, as a fraction of chord.
                 # Point at which the control surface is applied, as a fraction of chord.
                 control_surface_deflection: float = 0,  # Control deflection, in degrees. Downwards-positive.
                 spanwise_panels: int = 8,
                 # The number of spanwise panels to be used between this cross section and the next one.
                 spanwise_spacing: str = "cosine"  # Can be 'cosine' or 'uniform'. Highly recommended to be cosine.
                 ):
        if airfoil is None:
            raise ValueError("'airfoil' argument missing! (Needs an object of Airfoil type)")

        self.x_le = x_le
        self.y_le = y_le
        self.z_le = z_le

        self.chord = chord
        self.twist = twist
        self.twist_axis = twist_axis
        self.airfoil = airfoil
        self.control_surface_type = control_surface_type
        self.control_surface_hinge_point = control_surface_hinge_point
        self.control_surface_deflection = control_surface_deflection
        self.spanwise_panels = spanwise_panels
        self.spanwise_spacing = spanwise_spacing

        self.xyz_le = cas.vertcat(x_le, y_le, z_le)

    def __repr__(self) -> str:
        return "WingXSec (airfoil = %s, chord = %f, twist = %f)" % (
            self.airfoil.name,
            self.chord,
            self.twist
        )

    def quarter_chord(self) -> cas.DM:
        """
        Returns the (wing-relative) coordinates of the quarter chord of the cross section.
        """
        return 0.75 * self.xyz_le + 0.25 * self.xyz_te()

    def xyz_te(self) -> cas.DM:
        """
        Returns the (wing-relative) coordinates of the trailing edge of the cross section.
        """
        rot = angle_axis_rotation_matrix(self.twist * cas.pi / 180, self.twist_axis)
        xyz_te = self.xyz_le + rot @ cas.vertcat(self.chord, 0, 0)
        return xyz_te



import aerosandbox as asb
from aerosandbox import cas
import numpy as np
import pytest

"""
This test solves the N-dimensional Rosenbrock problem:

----------

Minimize:
sum(100*(x[i+1]-x[i]**2)**2 + (1-x[i])**2) for i in [1, N-1]

"""


def test_rosenbrock_constrained():
    opti = asb.Opti()  # set up an optimization environment

    N = 10  # Problem is unimodal for N=2, N=3, and N>=8. Bimodal for 4<=N<=7. Global min is always a vector of ones.

    x = opti.variable(N, init_guess=5)

    objective = 0
    for i in range(N - 1):
        objective += 100 * (x[i + 1] - x[i] ** 2) ** 2 + (1 - x[i]) ** 2

    opti.subject_to(x>=0) # Keeps design space unimodal for all N.

    opti.minimize(objective)

    sol = opti.solve()  # solve

    for i in range(N):
        assert sol.value(x[i]) == pytest.approx(1, abs=1e-4)


if __name__ == '__main__':
    pytest.main()



import numpy as np
import casadi as cas
from aerosandbox.tools.string_formatting import stdout_redirected
from aerosandbox.optimization.opti import Opti

def fit(
        model: callable,
        x_data: dict,
        y_data: np.ndarray,
        param_guesses: dict,
        param_bounds: dict = None,
        weights: np.ndarray = None,
        verbose: bool = True,
        scale_problem: bool = True,
        put_residuals_in_logspace: bool = False,
        residual_norm_type: str = "SSE",
):
    """
    Fits an analytical model to n datapoints using an automatic-differentiable optimization approach.
    :param model: A callable with syntax f(x, p) where:
            x is a dict of dependent variables. Same format as x_data [dict of 1D ndarrays of length n].
            p is a dict of parameters. Same format as param_guesses [dict of scalars].
        Model should return a 1D ndarray of length n.
        Model should use CasADi functions for differentiability.
    :param x_data: a dict of dependent variables. Same format as model's x. [1D ndarray or dict of 1D ndarrays of length n]
    :param y_data: independent variable. [1D ndarray of length n]
    :param param_guesses: a dict of fit parameters. Same format as model's p. Keys are parameter names, values are initial guesses. [dict of scalars]
    :param param_bounds: Optional: a dict of bounds on fit parameters.
        Keys are parameter names, values are a tuple of (min, max).
        May contain only a subset of param_guesses if desired.
        Use None to represent one-sided constraints (i.e. (None, 5)).
        [dict of tuples]
    :param weights: Optional: weights for data points. If not supplied, weights are assumed to be uniform.
        Weights are automatically normalized. [1D ndarray of length n]
    :param verbose: Whether or not to print information about parameters and goodness of fit.
    :param scale_problem: Whether or not to attempt to scale variables, constraints, and objective for more robust solve. [boolean]
    :param put_residuals_in_logspace: Whether to optimize using the logarithmic error as opposed to the absolute error (useful for minimizing percent error).
        Note: If any model outputs or data are negative, this will fail!
    :param residual_norm_type: What type of error norm should we use to optimize the fit parameters? [string]
        Options:
            * "SSE": minimizes the sum of squared errors.
            * "deviation": minimizes the maximum deviation of the fit; basically the infinity-norm of the error vector.
    :return: Optimal fit parameters [dict with same keys as param_guesses]
    """
    ### Initialize an optimization environment
    opti = Opti()

    ### Flatten all inputs
    def flatten(input):
        return np.array(input).flatten()

    try:
        x_data = {
            k: flatten(v)
            for k, v in x_data.items()
        }
    except AttributeError: # If it's not a dict, or a dict-like, put it in one
        x_data = {
            "x": flatten(x_data)
        }
    y_data = flatten(y_data)
    n_datapoints = len(y_data)

    ### Handle weighting
    if weights is None:
        weights = np.ones(n_datapoints)
    else:
        weights = flatten(weights)
    weights /= np.sum(weights)

    ### Check dimensionality of inputs to fitting algorithm
    for key, value in {
        **x_data,
        "y_data" : y_data,
        "weights": weights,
    }.items():
        # Check that the length of the inputs are consistent
        series_length = len(value)
        if not series_length == n_datapoints:
            raise ValueError(
                f"The supplied data series \"{key}\" has length {series_length}, but y_data has length {n_datapoints}.")

    def fit_param(initial_guess, lower_bound=None, upper_bound=None):
        """
        Helper function to create a fit variable
        :param initial_guess:
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        if scale_problem and np.abs(initial_guess) > 1e-8:
            var = initial_guess * opti.variable()  # scale variables
        else:
            var = opti.variable()
        opti.set_initial(var, initial_guess)
        if lower_bound is not None:
            lower_bound_abs = np.abs(lower_bound)
            if scale_problem and lower_bound_abs > 1e-8:
                opti.subject_to(var / lower_bound_abs > lower_bound / lower_bound_abs)
            else:
                opti.subject_to(var > lower_bound)
        if upper_bound is not None:
            upper_bound_abs = np.abs(upper_bound)
            if scale_problem and upper_bound_abs > 1e-8:
                opti.subject_to(var / upper_bound_abs < upper_bound / upper_bound_abs)
            else:
                opti.subject_to(var < upper_bound)
        return var

    if param_bounds is None:
        params = {
            k: fit_param(param_guesses[k])
            for k in param_guesses
        }
    else:
        params = {
            k: fit_param(param_guesses[k]) if k not in param_bounds else
            fit_param(param_guesses[k], param_bounds[k][0], param_bounds[k][1])
            for k in param_guesses
        }

    # if residual_type == "SSE":
    #     def error_norm(error_vector):
    #         return cas.sum1(weights * residuals ** 2)
    # elif:
    #     def error_norm(error_vector):
    #         return cas.fmax(residuals)
    # else:
    #     return ValueError("Bad input for the 'residual_type' parameter.")
    #

    def objective_function(params):
        """
        Given some parameters for the model, what is the "badness" of the corresponding fit.

        Args:
            params

        Returns: A scalar representing the "badness" of the fit.

        """
        y_model = model(x_data, params)
        if y_model is None:
            raise TypeError("model(x, param_guesses) returned None, when it should've returned a 1D ndarray.")

        if put_residuals_in_logspace:
            residuals = cas.log(y_model) - cas.log(y_data)
        else:
            residuals = y_model - y_data

        if residual_norm_type == "SSE":
            return cas.sum1(weights * residuals ** 2)
        elif residual_norm_type == "deviation":
            return cas.sum1(cas.fabs(weights * residuals))
        else:
            return ValueError("Bad input for the 'residual_type' parameter.")


    initial_objective = objective_function(param_guesses)

    if scale_problem:
        opti.minimize(objective_function(params)/initial_objective)
    else:
        opti.minimize(objective_function(params))

    # Solve
    if verbose:
        sol = opti.solve()
    else:
        with stdout_redirected():
            sol = opti.solve()

    params_solved = {}
    for k in params:
        try:
            params_solved[k] = sol.value(params[k])
        except:
            params_solved[k] = np.NaN

    # printing
    if verbose:
        # Print parameters
        print("\nFit Parameters:")
        if len(params_solved) <= 20:
            [print("\t%s = %f" % (k, v)) for k, v in params_solved.items()]
        else:
            print("\t%i parameters solved for." % len(params_solved))
        print("\nGoodness of Fit:")

        # Print objective function value
        print(f"\tInitial Badness (objective function): {initial_objective}")
        print(f"\tFinal Badness (objective function): {objective_function(params_solved)}")

        # Print R^2
        y_data_mean = cas.sum1(y_data) / y_data.shape[0]
        SS_tot = cas.sum1(weights * (y_data - y_data_mean) ** 2)
        SS_res = cas.sum1(weights * (y_data - model(x_data, params_solved)) ** 2)
        R_squared = sol.value(1 - SS_res / SS_tot)
        print(f"\tR^2: {R_squared}")

    return params_solved
